@@ -29,9 +29,7 @@ static void ftp_got_banner(struct connection *, struct read_buffer *);
 static void ftp_login(struct connection *);
 static void ftp_logged(struct connection *);
 static void ftp_sent_passwd(struct connection *);
-static void ftp_got_info(struct connection *, struct read_buffer *);
 static void ftp_got_user_info(struct connection *, struct read_buffer *);
-static void ftp_dummy_info(struct connection *, struct read_buffer *);
 static void ftp_pass_info(struct connection *, struct read_buffer *);
 static void ftp_send_retr_req(struct connection *, int, int);
 static struct ftp_connection_info *add_file_cmd_to_str(struct connection *, int);
@@ -39,7 +37,7 @@ static void ftp_retr_1(struct connection *);
 static void ftp_retr_file(struct connection *, struct read_buffer *);
 static void ftp_got_final_response(struct connection *, struct read_buffer *);
 static void created_data_connection(struct connection *);
-static void got_something_from_data_connection(struct connection *);
+static void got_something_from_data_connection(void *);
 static void ftp_end_request(struct connection *, int);
 static int get_ftp_response(struct connection *, struct read_buffer *, int);
 static int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int, int *);
@@ -80,8 +78,19 @@ static int get_ftp_response(struct connection *c, struct read_buffer *rb, int pa
 void ftp_func(struct connection *c)
 {
 	int we_are_in_root;
-	/*setcstate(c, S_CONN);*/
-	/*set_connection_timeout(c);*/
+	unsigned char *de, *d;
+	int del, bad_url;
+	d = get_url_data(c->url);
+	de = init_str(), del = 0;
+	add_conv_str(&de, &del, d, (int)strcspn(cast_const_char d, POST_CHAR_STRING), -2);
+	bad_url = !!strchr(cast_const_char de, 10);
+	mem_free(de);
+	if (bad_url) {
+		setcstate(c, S_BAD_URL);
+		abort_connection(c);
+		return;
+	}
+
 	if (get_keepalive_socket(c, &we_are_in_root)) {
 		int p;
 		if ((p = get_port(c->url)) == -1) {
@@ -89,7 +98,7 @@ void ftp_func(struct connection *c)
 			abort_connection(c);
 			return;
 		}
-		make_connection(c, p, &c->sock1, ftp_options.fast_ftp ? ftp_login : ftp_get_banner);
+		make_connection(c, p, &c->sock1, ftp_get_banner);
 	} else {
 		ftp_send_retr_req(c, S_SENT, we_are_in_root);
 	}
@@ -137,19 +146,7 @@ static void ftp_login(struct connection *c)
 	if ((u = get_user_name(c->url)) && *u) add_to_str(&login, &logl, u);
 	else add_to_str(&login, &logl, cast_uchar "anonymous");
 	if (u) mem_free(u);
-	if (ftp_options.fast_ftp) {
-		struct ftp_connection_info *fi;
-		add_to_str(&login, &logl, cast_uchar "\r\nPASS ");
-		u = get_ftp_password(c);
-		add_to_str(&login, &logl, u);
-		mem_free(u);
-		add_to_str(&login, &logl, cast_uchar "\r\n");
-		if (!(fi = add_file_cmd_to_str(c, 0))) {
-			mem_free(login);
-			return;
-		}
-		add_to_str(&login, &logl, fi->cmdbuf);
-	} else add_to_str(&login, &logl, cast_uchar "\r\n");
+	add_to_str(&login, &logl, cast_uchar "\r\n");
 	write_to_socket(c, c->sock1, login, logl, ftp_logged);
 	mem_free(login);
 	setcstate(c, S_SENT);
@@ -159,19 +156,6 @@ static void ftp_logged(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
-	if (!ftp_options.fast_ftp) {
-		ftp_got_user_info(c, rb);
-		return;
-	}
-	read_from_socket(c, c->sock1, rb, ftp_got_info);
-}
-
-static void ftp_got_info(struct connection *c, struct read_buffer *rb)
-{
-	int g = get_ftp_response(c, rb, 0);
-	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
-	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_info); return; }
-	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	ftp_got_user_info(c, rb);
 }
 
@@ -183,35 +167,23 @@ static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); retry_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	if (g >= 200 && g < 300) {
-		if (ftp_options.fast_ftp) ftp_dummy_info(c, rb);
-		else ftp_send_retr_req(c, S_GETH, 0);
+		ftp_send_retr_req(c, S_GETH, 0);
 	} else {
-		if (ftp_options.fast_ftp) ftp_pass_info(c, rb);
-		else {
-			unsigned char *login;
-			unsigned char *u;
-			int logl = 0;
-			login = init_str();
-			add_to_str(&login, &logl, cast_uchar "PASS ");
-			u = get_ftp_password(c);
-			add_to_str(&login, &logl, u);
-			mem_free(u);
-			add_to_str(&login, &logl, cast_uchar "\r\n");
-			write_to_socket(c, c->sock1, login, logl, ftp_sent_passwd);
-			mem_free(login);
-			setcstate(c, S_LOGIN);
-		}
+		unsigned char *login;
+		unsigned char *u;
+		int logl = 0;
+		login = init_str();
+		add_to_str(&login, &logl, cast_uchar "PASS ");
+		u = get_ftp_password(c);
+		add_to_str(&login, &logl, u);
+		mem_free(u);
+		add_to_str(&login, &logl, cast_uchar "\r\n");
+		write_to_socket(c, c->sock1, login, logl, ftp_sent_passwd);
+		mem_free(login);
+		setcstate(c, S_LOGIN);
 	}
 }
 
-static void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
-{
-	int g = get_ftp_response(c, rb, 0);
-	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
-	if (!g) { read_from_socket(c, c->sock1, rb, ftp_dummy_info); return; }
-	ftp_retr_file(c, rb);
-}
-
 static void ftp_sent_passwd(struct connection *c)
 {
 	struct read_buffer *rb;
@@ -226,8 +198,7 @@ static void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_pass_info); setcstate(c, S_LOGIN); return; }
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); abort_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); abort_connection(c); return; }
-	if (ftp_options.fast_ftp) ftp_retr_file(c, rb);
-	else ftp_send_retr_req(c, S_GETH, 0);
+	ftp_send_retr_req(c, S_GETH, 0);
 }
 
 static void add_port_pasv(unsigned char **s, int *l, struct ftp_connection_info *inf, unsigned char *port_string)
@@ -259,15 +230,9 @@ static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c, int
 	struct ftp_connection_info *inf, *inf2;
 	unsigned char *s;
 	int l;
-	if (!d) {
-		internal("get_url_data failed");
-		setcstate(c, S_INTERNAL);
-		abort_connection(c);
-		return NULL;
-	}
 
 	de = init_str(), del = 0;
-	add_conv_str(&de, &del, d, (int)strlen(cast_const_char d), -2);
+	add_conv_str(&de, &del, d, (int)strcspn(cast_const_char d, POST_CHAR_STRING), -2);
 	d = de;
 	inf = mem_alloc(sizeof(struct ftp_connection_info));
 	memset(inf, 0, sizeof(struct ftp_connection_info));
@@ -318,7 +283,7 @@ static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c, int
 #endif
 	dd = d;
 	while (*dd == '/') dd++;
-	if (!(de = cast_uchar strchr(cast_const_char dd, POST_CHAR))) de = cast_uchar strchr(cast_const_char dd, 0);
+	de = cast_uchar strchr(cast_const_char dd, 0);
 	if (dd == de || de[-1] == '/') {
 		inf->dir = 1;
 		inf->pending_commands = 3;
@@ -385,7 +350,7 @@ static void ftp_send_retr_req(struct connection *c, int state, int we_are_in_roo
 	} else {
 		fi = c->info;
 	}
-	if (ftp_options.fast_ftp) {
+	if (0) {
 		a:add_to_str(&login, &logl, fi->cmdbuf);
 	} else {
 		unsigned char *nl = cast_uchar strchr(cast_const_char fi->cmdbuf, '\n');
@@ -412,10 +377,8 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 	struct ftp_connection_info *inf = c->info;
 	if (0) {
 		rep:
-		if (!ftp_options.fast_ftp) {
-			ftp_send_retr_req(c, S_GETH, inf->we_are_in_root);
-			return;
-		}
+		ftp_send_retr_req(c, S_GETH, inf->we_are_in_root);
+		return;
 	}
 	if (inf->pending_commands > 1) {
 		unsigned char pc[6] = { 0, 0, 0, 0, 0, 0 }; /* against warning */
@@ -528,7 +491,7 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 		nol:;
 	}
 	if (!inf->pasv)
-		set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, c);
+		set_handlers(c->sock2, got_something_from_data_connection, NULL, c);
 	/*read_from_socket(c, c->sock1, rb, ftp_got_final_response);*/
 	ftp_got_final_response(c, rb);
 }
@@ -542,7 +505,7 @@ static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 	if (g == 425 || g == 450 || g == 500 || g == 501 || g == 550) {
 		if (c->url[strlen(cast_const_char c->url) - 1] == '/') goto skip_redir;
 		if (!c->cache) {
-			if (get_cache_entry(c->url, &c->cache)) {
+			if (get_connection_cache_entry(c)) {
 				setcstate(c, S_OUT_OF_MEM);
 				abort_connection(c);
 				return;
@@ -551,7 +514,6 @@ static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 		}
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
-		c->cache->redirect_get = 1;
 		add_to_strn(&c->cache->redirect, cast_uchar "/");
 		c->cache->incomplete = 0;
 		/*setcstate(c, S_FTP_NO_FILE);*/
@@ -764,11 +726,12 @@ static void created_data_connection(struct connection *c)
 	}
 #endif
 	inf->d = 1;
-	set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, c);
+	set_handlers(c->sock2, got_something_from_data_connection, NULL, c);
 }
 
-static void got_something_from_data_connection(struct connection *c)
+static void got_something_from_data_connection(void *c_)
 {
+	struct connection *c = (struct connection *)c_;
 	struct ftp_connection_info *inf = c->info;
 	int l;
 	int m;
@@ -783,11 +746,11 @@ static void got_something_from_data_connection(struct connection *c)
 		set_nonblock(ns);
 		EINTRLOOP(rs, close(c->sock2));
 		c->sock2 = ns;
-		set_handlers(ns, (void (*)(void *))got_something_from_data_connection, NULL, c);
+		set_handlers(ns, got_something_from_data_connection, NULL, c);
 		return;
 	}
 	if (!c->cache) {
-		if (get_cache_entry(c->url, &c->cache)) {
+		if (get_connection_cache_entry(c)) {
 			setcstate(c, S_OUT_OF_MEM);
 			abort_connection(c);
 			return;
@@ -799,14 +762,14 @@ static void got_something_from_data_connection(struct connection *c)
 		unsigned char *s0;
 		int s0l;
 		int err = 0;
-		static unsigned char ftp_head[] = "<html><head><title>/";
-		static unsigned char ftp_head2[] = "</title></head><body><h2>Directory /";
-		static unsigned char ftp_head3[] = "</h2><pre>";
+		static const unsigned char ftp_head[] = "<html><head><title>/";
+		static const unsigned char ftp_head2[] = "</title></head><body><h2>Directory /";
+		static const unsigned char ftp_head3[] = "</h2><pre>";
 #define A(s)							\
 do {								\
-	m = add_fragment(c->cache, c->from, s, strlen(cast_const_char s));	\
+	m = add_fragment(c->cache, c->from, s, strlen(cast_const_char s));\
 	if (m < 0 && !err) err = m;				\
-	c->from += strlen(cast_const_char s);					\
+	c->from += strlen(cast_const_char s);			\
 } while (0)
 		A(ftp_head);
 		ud = stracpy(get_url_data(c->url));
