@@ -10,6 +10,9 @@
 
 #define NUMBER_OF_DEVICES	12
 
+static void grx_mouse_init(void);
+static void grx_mouse_terminate(void);
+
 void _GrCloseVideoDriver(void);
 
 extern struct graphics_driver grx_driver;
@@ -25,23 +28,43 @@ static unsigned char in_graphics_mode = 0;
 
 static void grx_set_palette(void)
 {
-	int i;
-	if (grx_driver.depth == 33) {
-		for (i = 0; i < 16; i++) {
-			GrSetColor(i,
-				i & 8 ? 255 : 0,
-				((i >> 1) & 3) * 85,
-				i & 1 ? 255 : 0);
-		}
-	}
-	if (grx_driver.depth == 65) {
-		for (i = 0; i < 256; i++) {
-			GrSetColor(i,
-				((i >> 5) & 7) * (255./7.),
-				((i >> 2) & 7) * (255./7.),
-				(i & 3) * 85);
-		}
+	unsigned i, colors;
+	if (grx_driver.depth == 33)
+		colors = 16;
+	else if (grx_driver.depth == (65 | 0x300))
+		colors = 216;
+	else if (grx_driver.depth == 65)
+		colors = 256;
+	else
+		return;
+	GrResetColors();
+	for (i = 0; i < colors; i++) {
+		unsigned rgb[3];
+		q_palette(colors, i, 255, rgb);
+		GrSetColor(i, rgb[0], rgb[1], rgb[2]);
 	}
+	return;
+}
+
+static void grx_update_palette(void)
+{
+	int new_depth;
+	if (!grx_driver.param->palette_mode)
+		new_depth = grx_driver.depth | 0x300;
+	else
+		new_depth = grx_driver.depth & ~0x300;
+
+	if (new_depth == grx_driver.depth)
+		return;
+
+	grx_driver.depth = new_depth;
+
+	grx_driver.get_color = get_color_fn(new_depth);
+	if (!grx_driver.get_color) internal("Unknown bit depth %x", new_depth);
+
+	grx_mouse_terminate();
+	grx_set_palette();
+	grx_mouse_init();
 }
 
 static void grx_set_text(void)
@@ -79,7 +102,7 @@ static void grx_key_in(struct itrm *p, unsigned char *ev_, int size)
 	if (size != sizeof(struct links_event)) return;
 	if (ev->ev == EV_ABORT) terminate_loop = 1;
 	if (ev->ev != EV_KBD) return;
-	if (ev->y & KBD_PASTE) goto skip;
+	if (ev->y & KBD_PASTING) goto skip;
 	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
 		vd = (ev->x - '1' + 10) % 10;
 		goto switch_vd;
@@ -134,7 +157,7 @@ static void grx_mouse_init(void)
 
 	grx_mouse_cursor = GrBuildCursor(cast_char cursor, arrow_width, arrow_width, arrow_height, 0, 0, table);
 
-	if (!grx_mouse_cursor) GrMouseSetColors(GrWhite(), GrBlack());
+	if (!grx_mouse_cursor) GrMouseSetColors(grx_driver.get_color(0xffffff), grx_driver.get_color(0x000000));
 	else GrMouseSetCursor(grx_mouse_cursor);
 
 	GrMouseDisplayCursor();
@@ -143,6 +166,7 @@ static void grx_mouse_init(void)
 static void grx_mouse_terminate(void)
 {
 	if (!grx_mouse_initialized) return;
+	GrMouseEraseCursor();
 	GrMouseUnInit();
 	if (grx_mouse_cursor) GrDestroyCursor(grx_mouse_cursor);
 	grx_mouse_initialized = 0;
@@ -194,9 +218,11 @@ static unsigned char *set_mode_according_to_param(unsigned char *param)
 #endif
 	if (!param || !*param) {
 		return stracpy(cast_uchar "There is no default videomode.\nUse the switch \"-mode WIDTHxHEIGHTxCOLORS\" to select video mode.\nThe COLORS argument may be 16, 256, 32k, 64k, 16M.\n");
+#if 0
 		if (!GrSetMode(GR_default_graphics))
 			return stracpy(cast_uchar "GrSetMode failed.\n");
 		in_graphics_mode = 1;
+#endif
 	} else {
 		unsigned long x, y;
 		int pl;
@@ -261,12 +287,16 @@ unsupported_videomode:
 		grx_restore();
 		return stracpy(e);
 	}
+	grx_driver.flags &= ~GD_SELECT_PALETTE;
 	switch (mode->bpp) {
 		case 4:
 			grx_driver.depth = 33;
 			break;
 		case 8:
 			grx_driver.depth = 65;
+			if (!grx_driver.param->palette_mode)
+				grx_driver.depth |= 0x300;
+			grx_driver.flags |= GD_SELECT_PALETTE;
 			break;
 		case 15:
 			grx_driver.depth = 122;
@@ -287,6 +317,7 @@ unsupported_videomode:
 			goto unsupported_videomode;
 	}
 	grx_driver.get_color = get_color_fn(grx_driver.depth);
+	if (!grx_driver.get_color) internal("Unknown bit depth %x", grx_driver.depth);
 	grx_driver.x = mode->width;
 	grx_driver.y = mode->height;
 	if (mode->bpp >= 8) grx_driver.flags |= GD_DONT_USE_SCROLL;
@@ -297,10 +328,7 @@ unsupported_videomode:
 
 	grx_driver_param = stracpy(param ? param : cast_uchar "");
 
-	if (init_virtual_devices(&grx_driver, NUMBER_OF_DEVICES)) {
-		grx_restore();
-		return stracpy(cast_uchar "Allocation of virtual devices failed.\n");
-	}
+	init_virtual_devices(&grx_driver, NUMBER_OF_DEVICES);
 
 	svgalib_kbd = handle_svgalib_keyboard(grx_key_in);
 	grx_mouse_init();
@@ -432,6 +460,7 @@ static void grx_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int
 	struct grx_bmp *b = bmp->flags;
 	if (!b) return;
 	TEST_INACTIVITY
+	CLIP_DRAW_BITMAP
 	/* note: GrImageDisplay has buggy clipping.
 	   It doesn't draw if the target x or y coordinate is one pixel
 	   before the end of clip area */
@@ -441,6 +470,7 @@ static void grx_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int
 static void grx_fill_area(struct graphics_device *dev, int x1, int y1, int x2, int y2, long color)
 {
 	TEST_INACTIVITY
+	CLIP_FILL_AREA
 	if (x1 >= x2 || y1 >= y2) return;
 	GrFilledBox(x1, y1, x2 - 1, y2 - 1, color);
 }
@@ -448,56 +478,34 @@ static void grx_fill_area(struct graphics_device *dev, int x1, int y1, int x2, i
 static void grx_draw_hline(struct graphics_device *dev, int x1, int y, int x2, long color)
 {
 	TEST_INACTIVITY
-	if (x1 >= x2) return;
+	CLIP_DRAW_HLINE
 	GrHLine(x1, x2 - 1, y, color);
 }
 
 static void grx_draw_vline(struct graphics_device *dev, int x, int y1, int y2, long color)
 {
 	TEST_INACTIVITY
-	if (y1 >= y2) return;
+	CLIP_DRAW_VLINE
 	GrVLine(x, y1, y2 - 1, color);
 }
 
-static int grx_hscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
+static int grx_scroll(struct graphics_device *dev, struct rect_set **ignore, int scx, int scy)
 {
 	TEST_INACTIVITY_0
-	if (!sc) return 0;
-	if (sc < 0) {
-		if (-sc >= dev->clip.x2 - dev->clip.x1) return 1;
-		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1,
-			NULL, dev->clip.x1 - sc, dev->clip.y1,
-			dev->clip.x2 - 1, dev->clip.y2 - 1, GrWRITE);
-	} else {
-		if (sc >= dev->clip.x2 - dev->clip.x1) return 1;
-		GrBitBlt(NULL, dev->clip.x1 + sc, dev->clip.y1,
-			NULL, dev->clip.x1, dev->clip.y1,
-			dev->clip.x2 - sc - 1, dev->clip.y2 - 1, GrWRITE);
-	}
+	GrBitBlt(NULL,
+		dev->clip.x1 + (scx >= 0 ? scx : 0),
+		dev->clip.y1 + (scy >= 0 ? scy : 0),
+		NULL,
+		dev->clip.x1 - (scx < 0 ? scx : 0),
+		dev->clip.y1 - (scy < 0 ? scy : 0),
+		dev->clip.x2 - (scx >= 0 ? scx : 0) - 1,
+		dev->clip.y2 - (scy >= 0 ? scy : 0) - 1,
+		GrWRITE);
 	return 1;
 }
 
-static int grx_vscroll(struct graphics_device *dev, struct rect_set **ignore, int sc)
+static void grx_set_clip_area(struct graphics_device *dev)
 {
-	TEST_INACTIVITY_0
-	if (!sc) return 0;
-	if (sc < 0) {
-		if (-sc >= dev->clip.y2 - dev->clip.y1) return 1;
-		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1,
-			NULL, dev->clip.x1, dev->clip.y1 - sc,
-			dev->clip.x2 - 1, dev->clip.y2 - 1, GrWRITE);
-	} else {
-		if (sc >= dev->clip.y2 - dev->clip.y1) return 1;
-		GrBitBlt(NULL, dev->clip.x1, dev->clip.y1 + sc,
-			NULL, dev->clip.x1, dev->clip.y1,
-			dev->clip.x2 - 1, dev->clip.y2 - sc - 1, GrWRITE);
-	}
-	return 1;
-}
-
-static void grx_set_clip_area(struct graphics_device *dev, struct rect *r)
-{
-	generic_set_clip_area(dev, r);
 	if (dev != current_virtual_device) return;
 	grx_set_clip();
 }
@@ -552,12 +560,12 @@ struct graphics_driver grx_driver = {
 	grx_fill_area,
 	grx_draw_hline,
 	grx_draw_vline,
-	grx_hscroll,
-	grx_vscroll,
+	grx_scroll,
 	grx_set_clip_area,
 	NULL,
 	grx_block,
 	grx_unblock,
+	grx_update_palette, /* set_palette */
 	NULL, /* set_title */
 	NULL, /* exec */
 	NULL, /* set_clipboard_text */
@@ -565,8 +573,7 @@ struct graphics_driver grx_driver = {
 	0,				/* depth */
 	0, 0,				/* size */
 	GD_NEED_CODEPAGE,		/* flags */
-	0,				/* codepage */
-	NULL,				/* shell */
+	NULL,				/* param */
 };
 
 #endif
