@@ -149,6 +149,7 @@ static int check_http_server_bugs(unsigned char *url, struct http_connection_inf
 
 static void http_end_request(struct connection *c, int notrunc, int nokeepalive, int state)
 {
+	struct http_connection_info *info = c->info;
 	if (state == S__OK) {
 		if (c->cache) {
 			if (!notrunc) truncate_entry(c->cache, c->from, 1);
@@ -157,8 +158,8 @@ static void http_end_request(struct connection *c, int notrunc, int nokeepalive,
 	}
 	setcstate(c, state);
 	if (c->info &&
-	    !((struct http_connection_info *)c->info)->close &&
-	    !((struct http_connection_info *)c->info)->send_close &&
+	    !info->close &&
+	    !info->send_close &&
 	    !nokeepalive) {
 		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT, 0);
 	} else {
@@ -245,7 +246,7 @@ static void http_send_header(struct connection *c)
 		post = cast_uchar strchr(cast_const_char host, POST_CHAR);
 		if (post) post++;
 	}
-	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive);
+	info->send_close = info->https_forward || http10;
 #ifdef HAVE_SSL
 	if (info->https_forward) {
 		add_to_str(&hdr, &l, cast_uchar "CONNECT ");
@@ -773,7 +774,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 		if (info->length >= 0) info->length -= l;
 		c->from += l;
 		kill_buffer_data(rb, l);
-		if (!info->length && !rb->close) {
+		if (!info->length) {
 			http_end_request(c, 0, 0, S__OK);
 			return;
 		}
@@ -964,7 +965,6 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		unsigned char *cookie;
 		unsigned char *ch = head;
 		while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-			unsigned char *host = remove_proxy_prefix(c->url);
 			set_cookie(NULL, host, cookie);
 			mem_free(cookie);
 		}
@@ -1060,7 +1060,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 #endif
 	if (e->redirect) mem_free(e->redirect), e->redirect = NULL;
 	if ((h == 302 || h == 303 || h == 307 || h == 511) && !e->expire_time) e->expire_time = 1;
-	if (h == 301 || h == 302 || h == 303 || h == 307) {
+	if (h == 301 || h == 302 || h == 303 || h == 307 || h == 308) {
 		if ((d = parse_http_header(e->head, cast_uchar "Location", NULL))) {
 			unsigned char *user, *ins;
 			unsigned char *newuser, *newpassword;
@@ -1082,7 +1082,11 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (e->redirect) mem_free(e->redirect);
 			e->redirect = d;
-			e->redirect_get = h == 303;
+			if (h == 307 || h == 308) {
+				unsigned char *p;
+				if ((p = cast_uchar strchr(cast_const_char host, POST_CHAR)))
+					add_to_strn(&e->redirect, p);
+			}
 		}
 	}
 	if (!e->expire_time && strchr(cast_const_char c->url, POST_CHAR)) e->expire_time = 1;
@@ -1118,7 +1122,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		unsigned char *ep;
 		my_strtoll_t l = my_strtoll(d, &ep);
 		if (!*ep && l >= 0 && (off_t)l >= 0 && (off_t)l == l) {
-			if (!info->close || version >= 11) info->length = l;
+			if (!info->close || version >= 11 || h / 100 == 3) info->length = l;
 			if (c->from + l >= 0) c->est_length = c->from + l;
 		}
 		mem_free(d);
